#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "driverlib/sysctl.h"
#include "driverlib/gpio.h"
#include "driverlib/debug.h"
#include "driverlib/pin_map.h"
#include "inc/hw_gpio.h"
#include "driverlib/rom.h"
#include "driverlib/timer.h"
#include "driverlib/interrupt.h"
#include "driverlib/uart.h"
#include "inc/hw_ints.h"
#include "driverlib/interrupt.h"
#include "utils/uartstdio.h"

typedef char *string;

#define MAX_BUTTONS 5
int TimerFlag = 0;
//void (*Timer0ISR) (void);

void UARTInit (void)
{
	SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOB);
	GPIOPinConfigure(GPIO_PB0_U1RX);
	GPIOPinConfigure(GPIO_PB1_U1TX);
	GPIOPinTypeUART(GPIO_PORTB_BASE, GPIO_PIN_0 | GPIO_PIN_1);
	UARTStdioConfig(1, 9600, SysCtlClockGet());
	UARTEchoSet(false);
}

void GPIOInit()
{
	SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOC);
	GPIODirModeSet(GPIO_PORTC_BASE, GPIO_PIN_4|GPIO_PIN_5|GPIO_PIN_6|GPIO_PIN_7, GPIO_DIR_MODE_IN);
	GPIOPadConfigSet(GPIO_PORTC_BASE, GPIO_PIN_4|GPIO_PIN_5|GPIO_PIN_6|GPIO_PIN_7, GPIO_STRENGTH_2MA, GPIO_PIN_TYPE_STD_WPD);
	SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOB);
	GPIODirModeSet(GPIO_PORTB_BASE, GPIO_PIN_3, GPIO_DIR_MODE_IN);
	GPIOPadConfigSet(GPIO_PORTB_BASE, GPIO_PIN_3, GPIO_STRENGTH_2MA, GPIO_PIN_TYPE_STD_WPD);


}

void TimerInit()
{
	uint32_t ui32Period;
	IntMasterDisable();
	SysCtlPeripheralEnable(SYSCTL_PERIPH_TIMER0);
	TimerDisable(TIMER0_BASE, TIMER_A);
	TimerIntClear(TIMER0_BASE, TIMER_TIMA_TIMEOUT);
	TimerConfigure(TIMER0_BASE, TIMER_CFG_PERIODIC);
	ui32Period = (SysCtlClockGet() / 10) / 2;		//10 interrupts a second
	TimerLoadSet(TIMER0_BASE, TIMER_A, ui32Period -1);
	//TimerIntRegister(TIMER0_BASE, TIMER_A, Timer0ISR);
	IntEnable(INT_TIMER0A);
	TimerIntEnable(TIMER0_BASE, TIMER_TIMA_TIMEOUT);
}



int ReadButton (int i)
{
	switch(i)
	{
	case 0:
			return GPIOPinRead(GPIO_PORTB_BASE, GPIO_PIN_3);		//STOP PB3
	case 1:
			return GPIOPinRead(GPIO_PORTC_BASE, GPIO_PIN_4);		//FORWARD PC4
	case 2:
			return GPIOPinRead(GPIO_PORTC_BASE, GPIO_PIN_5);		//REVERSE PC5
	case 3:
			return GPIOPinRead(GPIO_PORTC_BASE, GPIO_PIN_6);		//LEFT PC6
	case 4:
			return GPIOPinRead(GPIO_PORTC_BASE, GPIO_PIN_7);		//RIGHT PC7
	}
	return -1;
}
//Timer ISR routine take a pointer to the DebouncedButtons array, return nothing,
//Button_Count has to remain static as well, declare in header?

void Timer0ISR (void)
{
	TimerIntClear(TIMER0_BASE, TIMER_A);
	TimerFlag = 1;
	TimerDisable(TIMER0_BASE, TIMER_A);
}

void Debounce (int *Button_Count, int *DebouncedButtons)
{
	int i;
	for (i=0;i<MAX_BUTTONS;i++)
			{
				if (ReadButton(i))
				{
					if(Button_Count[i] <6)
					{
						Button_Count[i]++;
					}
					if(Button_Count[i] >5)
					{
						DebouncedButtons[i] = 1;
					}
				}else
				{
					if(Button_Count[i] > 0)
						{
						Button_Count[i]--;
						}
					if (Button_Count[i]<3)
					{
						DebouncedButtons[i] = 0;
					}
				}
			}
}

string DecipherButtons (int *DebouncedButtons)
{
	int i;
	for(i=0;i<MAX_BUTTONS;i++)
	{
		if(DebouncedButtons[i] == 1)
		{
			if(DebouncedButtons[0] == 1)
			{
				return "Stop";
			}else
			{
				if(DebouncedButtons[1]&&DebouncedButtons[2] || DebouncedButtons[3]&&DebouncedButtons[4])
				{
					return "illegal";
				}else if(DebouncedButtons[1] == 1)
				{
					if(DebouncedButtons[3] == 1)
					{
						return "ForLeft";
					}else if(DebouncedButtons[4] == 1)
					{
						return "ForRight";
					}else
					{
						return "Forward";
					}
				}else if(DebouncedButtons[2] == 1)
				{
					if(DebouncedButtons[3] == 1)
					{
						return "RevLeft";
					}else if(DebouncedButtons[4] == 1)
					{
						return "RevRight";
					}else
					{
						return "Reverse";
					}
				}else if(DebouncedButtons[3] == 1)
				{
					return "Left";
				}else if(DebouncedButtons[4] == 1)
				{
					return "Right";
				}
			}
		}
	}
	return "NoButton";
}
void Comms (char *Button)
{

	/*char RXstring1[17];
	char Command[17] = "1234567890123456";
	SysCtlDelay(3200);
	UARTgets(RXstring1, 16);
	if(strcmp(RXstring1, "") != 0)
	{
	strcpy(Command,RXstring1);
	}
	if (strcmp(Command, "command") == 0)
	{*/
		UARTprintf("%s\n", Button);
		/*UARTFlushRx();
	}else
	{
		UARTprintf("NOTVALID\n");
		UARTFlushRx();
	}*/
}

void main(void) {
	int DebouncedButtons[MAX_BUTTONS] = {0,0,0,0,0};	//array of int's 1 cell for each button, 0 means un-pressed, 1 means pressed, some other number means holy shit we fucked up
	int Button_Count[MAX_BUTTONS] = {0,0,0,0,0};		//Counts/votes on samples that a button is pressed. needs to stay static.
	char ButtonPrev[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	string Button = "Stop";
	GPIOInit();
	TimerInit();
	UARTInit();
	IntMasterEnable();
	TimerEnable(TIMER0_BASE, TIMER_A);

	while(1)
	{
		TimerFlag = 0;
		Debounce(Button_Count, DebouncedButtons);
		strcpy(ButtonPrev, Button);
		Button = DecipherButtons(DebouncedButtons);
		if (strcmp(Button, ButtonPrev) != 0)
		{
			Comms(Button);
		}
		/*if(UARTRxBytesAvail() != 0)
		{
			SysCtlDelay(SysCtlClockGet()*9/9600);
			if(UARTPeek('\r') > 0)
			{
			Comms(Button);
			}else
			{
				UARTFlushRx();
			}
		}*/

	}
}
