#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "driverlib/sysctl.h"
#include "driverlib/gpio.h"
#include "driverlib/debug.h"
#include "inc/hw_ints.h"
#include "driverlib/interrupt.h"
#include "driverlib/pwm.h"
#include "driverlib/pin_map.h"
#include "inc/hw_gpio.h"
#include "driverlib/rom.h"
#include "driverlib/qei.h"
#include "driverlib/adc.h"
#include "driverlib/uart.h"
#include "utils/uartstdio.h"
#include "utils/uartstdio1.h"


typedef char *string;
struct Speed {
			int Master;
			int Slave;
		};
#define PWM_FREQUENCY 50
//*******************************************************************
//
// 	These are values you can change to fine tune the performance of
//  your stander.
//	Change TOGGLE to 1 if you want the buttons to exhibit Toggling
//	behavior, where a button will stay pressed until it is pressed
//	again or another button is pressed
//	TRIMF will increase the slave wheel speed in comparison with the
//	master wheel going forward
//	TRIMR will do the same in reverse
//********************************************************************
#define TOGGLE 0
#define KI 25
#define KP 15
#define TRIMF 5
#define TRIMR 5

//*******************************************************************
//
//	These are all the initializations, you should only mess with
//	line 100 and 101, unless you know what you are doing
//
//*******************************************************************

void PWMInit()
{
		volatile int ui32Load;
		volatile int ui32PWMClock;
		volatile int ui32Width;
		volatile int ui8Adjust;

		ui8Adjust = 150;
		SysCtlPWMClockSet(SYSCTL_PWMDIV_64);
		SysCtlPeripheralEnable(SYSCTL_PERIPH_PWM0);
		SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOD);
		GPIOPinTypePWM(GPIO_PORTD_BASE, GPIO_PIN_0|GPIO_PIN_1);
		GPIOPinConfigure(GPIO_PD0_M0PWM6);
		GPIOPinConfigure(GPIO_PD1_M0PWM7);
		ui32PWMClock = SysCtlClockGet() / 64;
		ui32Load = (ui32PWMClock / PWM_FREQUENCY) - 1;
		ui32Width = (ui8Adjust * ui32Load / 2000);
		PWMGenConfigure(PWM0_BASE, PWM_GEN_3, PWM_GEN_MODE_DOWN);
		PWMGenPeriodSet(PWM0_BASE, PWM_GEN_3, ui32Load);
		PWMPulseWidthSet(PWM0_BASE, PWM_OUT_6, ui32Width);
		PWMPulseWidthSet(PWM0_BASE, PWM_OUT_7, ui32Width);
		PWMOutputState(PWM0_BASE, PWM_OUT_6_BIT|PWM_OUT_7_BIT, true);
		PWMGenEnable(PWM0_BASE, PWM_GEN_3);
}

void GPIOInit()		//Enables Pin B4 as an active high pin for the bumper
{
	SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOB);
	GPIODirModeSet(GPIO_PORTB_BASE, GPIO_PIN_4, GPIO_DIR_MODE_IN);
	GPIOPadConfigSet(GPIO_PORTB_BASE, GPIO_PIN_4, GPIO_STRENGTH_2MA, GPIO_PIN_TYPE_STD_WPD);
}
void QEIInit()
	{
		SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOD);
		SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOC);
		SysCtlPeripheralEnable(SYSCTL_PERIPH_QEI0);
		SysCtlPeripheralEnable(SYSCTL_PERIPH_QEI1);
		//Unlock GPIOD7 - Like PF0 its used for NMI - Without this step it doesn't work
			HWREG(GPIO_PORTD_BASE + GPIO_O_LOCK) = GPIO_LOCK_KEY; //In Tiva include this is the same as "_DD" in older versions (0x4C4F434B)
			HWREG(GPIO_PORTD_BASE + GPIO_O_CR) |= 0x80;
			HWREG(GPIO_PORTD_BASE + GPIO_O_LOCK) = 0;
		GPIOPinConfigure(GPIO_PD6_PHA0);
		GPIOPinConfigure(GPIO_PD7_PHB0);
		GPIOPinConfigure(GPIO_PC5_PHA1);
		GPIOPinConfigure(GPIO_PC6_PHB1);
		GPIOPinTypeQEI(GPIO_PORTD_BASE, GPIO_PIN_6|GPIO_PIN_7);
		GPIOPinTypeQEI(GPIO_PORTC_BASE, GPIO_PIN_5|GPIO_PIN_6);
		QEIDisable(QEI0_BASE);
		QEIDisable(QEI1_BASE);
		QEIIntDisable(QEI0_BASE, QEI_INTERROR|QEI_INTDIR|QEI_INTTIMER|QEI_INTINDEX);
		QEIIntDisable(QEI1_BASE, QEI_INTERROR|QEI_INTDIR|QEI_INTTIMER|QEI_INTINDEX);
		QEIConfigure(QEI0_BASE, QEI_CONFIG_CAPTURE_A_B|QEI_CONFIG_NO_RESET|QEI_CONFIG_QUADRATURE|QEI_CONFIG_SWAP, 3000);	//the MAX_POSITION included at the end of the function call is not used here because of "QEI_CONFIG_NO_RESET"
		QEIConfigure(QEI1_BASE, QEI_CONFIG_CAPTURE_A_B|QEI_CONFIG_NO_RESET|QEI_CONFIG_QUADRATURE|QEI_CONFIG_NO_SWAP, 3000);
		QEIEnable(QEI0_BASE);
		QEIEnable(QEI1_BASE);
		QEIVelocityConfigure(QEI0_BASE, QEI_VELDIV_1, 16);
		QEIVelocityEnable(QEI0_BASE);
		QEIPositionSet(QEI0_BASE, 1500);
		QEIPositionSet(QEI1_BASE, 1500);
	}

void ADCInit()
{
	//Emable Pin PE2
	SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOE);
	GPIOPinTypeADC(GPIO_PORTE_BASE, GPIO_PIN_2);
	//Enable ADC0 peripheral
	SysCtlPeripheralEnable(SYSCTL_PERIPH_ADC0);

	// Now, this is an API call that specifies the samples to be averaged - 2, 4, 8, 32, or 64.
	//Each sample is the results of 64 averaged measurements.
	//ADCHardwareOversampleConfigure(ADC0_BASE, 4);

	//NOTE: If ADC12 not needed to run at 1Msps, need to configure an extra portion.

	// Here they are using the ADC sequencer with the highest priority.
	ADCSequenceConfigure(ADC0_BASE, 1, ADC_TRIGGER_PROCESSOR, 0);

	//Here they are configuring the ADC Sequencer's 4 steps.
	// Steps 0-2 of sequencer 1 sample the temperature sensor (ADC_CTL_TS)
	// In this example they'll sample all 4 temperature samples to calculate temp value.
	ADCSequenceStepConfigure(ADC0_BASE, 1, 0, ADC_CTL_CH1);
	ADCSequenceStepConfigure(ADC0_BASE, 1, 1, ADC_CTL_CH1);
	ADCSequenceStepConfigure(ADC0_BASE, 1, 2, ADC_CTL_CH1);

	//Final sequencer step samples the temp sensor and configures the interrupt flag to pop up when the sampling is finished.
	ADCSequenceStepConfigure(ADC0_BASE, 1, 0, ADC_CTL_CH1|ADC_CTL_IE|ADC_CTL_END);
	//And thus, now ADC Sequencer 1 can be enabled:
	ADCSequenceEnable(ADC0_BASE, 1);
}

void UARTInit (void)
{
		SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOA);
		GPIOPinConfigure(GPIO_PA0_U0RX);
		GPIOPinConfigure(GPIO_PA1_U0TX);
		GPIOPinTypeUART(GPIO_PORTA_BASE, GPIO_PIN_0 | GPIO_PIN_1);
		UARTStdioConfig(0, 9600, SysCtlClockGet());
		//Enable UARTStdio1 to communicate with remote
		SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOE);
		GPIOPinConfigure(GPIO_PE4_U5RX);
		GPIOPinConfigure(GPIO_PE5_U5TX);
		GPIOPinTypeUART(GPIO_PORTE_BASE, GPIO_PIN_4|GPIO_PIN_5);
		UART1StdioConfig(5,9600, SysCtlClockGet());
		UARTEchoSet(false);
		UART1EchoSet(false);
		UARTFlushRx();
		UART1FlushRx();
}
//*********************************************************
//
//	Speed get is supposed to get the speed from the ADC.
//	Unfortunately, it doesn't work. I might have mixed up
//	the pin to it. Or I might have done the Init wrong.
//
//*********************************************************
int SpeedGet()
{
	//This creates an array to store the ADC Data. - Look at FIFO characteristics for more details
	uint32_t ui32ADC0Value[4];

	//These variables store calculated Speed values from the sensor. They are volatile .:. they are not optimized by compiler and are available in local window at run-time.
	volatile uint32_t ui32Speed;
	//Always clear the interrupt before starting the program.
			ADCIntClear(ADC0_BASE, 1);
			//When the int flag actually pops up, it will tell us that the sampling has been completed.

			//Now we'll trigger the ADC Conversion via software:
			ADCProcessorTrigger(ADC0_BASE, 1);

			//The conversion must be completed, probably a good idea to use an interrupt - here they just waste CPU cycles.
			while(!ADCIntStatus(ADC0_BASE, 1, false))
			{
			}

		ADCSequenceDataGet(ADC0_BASE, 1, ui32ADC0Value);

		//Calculate the average temp sensor data. (Floating point operations later - this is fixed-point)
		 ui32Speed = (ui32ADC0Value[0] + ui32ADC0Value[1] + ui32ADC0Value[2] + ui32ADC0Value[3] + 2)/4;
		 return ui32Speed;
}
//***********************************************************
//	LookupControl turns a string command such as "Forward"
//	into a number which can be used in the switch statements
//	that follow it.
//
//***********************************************************

int LookupControl (char *Command)
{
	int i=0;
	int Direction;
	string CommandList[] =
	{
	            "Stop",
				"Forward",
				"Reverse",
				"Left",
				"Right",
				"ForLeft",
				"ForRight",
				"RevLeft",
				"RevRight",
	};
	for (i=0; i<8; i++)
	{
		if (strcmp(CommandList[i], Command) == 0)
		{
			Direction = i;
		}
	}
	return Direction;
}
//********************************************************
//
//	ChangeSpeed is called when a button is pushed and you
//	want to change the target speed you ant to go. The
//	max speed is set here on line 247 with ADCSpeed.
//	If you get the ADC working, then you can make it a
//	function call
//
//********************************************************

void ChangeSpeed (char *Command, int ui32Load, struct Speed *Target)
{
	struct Speed CurrentSpeed;
	volatile int ADCSpeed;
	volatile int TargetDirection;

	CurrentSpeed.Master = PWMPulseWidthGet(PWM0_BASE, PWM_OUT_6);
	CurrentSpeed.Master = CurrentSpeed.Slave *2000/ui32Load;			// This calculates the ui8Adjust values (why am I not just passing them in?)
	CurrentSpeed.Slave = PWMPulseWidthGet(PWM0_BASE, PWM_OUT_7);
	CurrentSpeed.Slave = CurrentSpeed.Master *2000/ui32Load;
	//QEIPositionSet(QEI0_BASE, 1500);
	//QEIPositionSet(QEI1_BASE, 1500);
	ADCSpeed = 20; //SpeedGet()/1027*35;
	TargetDirection = LookupControl(Command);
	switch(TargetDirection)
	{
	case 0:		//Stop
		Target->Master = 150;
		Target->Slave = 150;
		break;

	case 1:		//Forward
		Target->Master = 150 + (ADCSpeed);
		Target->Slave = Target->Master;
		if(Target->Slave > 170)
			{
			 Target->Slave = 170;
			 Target->Master = 170;
			}
		break;
	case 2:		//Reverse
		Target->Master = 150 - (ADCSpeed);//*2/3);
		Target->Slave = Target->Master;
		if(Target->Slave < 130)
			{
			 Target->Slave = 130;
			 Target->Master = 130;
			}
		break;
	case 3: 	//Left
		Target->Master = 150 + (ADCSpeed);//*2/3);
		Target->Slave = 150;
		break;
	case 4:		//Right
		Target->Master = 150;
		Target->Slave = 150 + (ADCSpeed);//*2/3);
		break;
	case 5: 	//Forward Left
		Target->Master = 150 + (ADCSpeed);//*2/3);
		Target->Slave = 150 + (ADCSpeed*2/3);
		break;
	case 6:		//Forward Right
		Target->Master = 150 + (ADCSpeed*2/3);
		Target->Slave = 150 + (ADCSpeed);//*2/3);
		break;
	case 7:		//Reverse Left
		Target->Master = 150 - (ADCSpeed);//*2/3);
		Target->Slave = 150 - (ADCSpeed*2/3);
		break;
	case 8:		//Reverse Right
		Target->Master = 150 - (ADCSpeed*2/3);
		Target->Slave = 150 - (ADCSpeed);
	break;
	default:
		Target->Master = 150;
		Target->Slave = 150;
		break;
	}

}

//************************************************
//
//	WheelMonitorLoop is called to change the
//	voltage to the wheels. It is responsible for
//	the acceleration, and straight travel of the
//	stander. There are 3 large portions of this
//	function.
//	1) Keep the position counters in the middle of
//	their range by setting the master position
//	to the middle of its range and subtracting
//	the difference in position from the slave
//	position. If the position difference is too
//	large, it slices the difference to 1/3
//	2)Set a target speed for this loop. This is
//	responsible for the acceleration of that stander
//	Its not noticeable on the stander because the
//	acceleration value is too large for the time of
//	the loop
//	3) The straight driving closed loop control. The
//	difference in position is the error to the slave
//	wheel. The Integral term is commented out
//
//*************************************************
void WheelMonitorLoop(struct Speed *Target, struct Speed *NewSpeed, struct Speed *Uncorrected,  uint32_t ui32Load, char *Button, int *Integral)
{
	int a = 1;	//Acceleration constant. maximum amount speed will change by in one cycle
	int TargetDirection;
	int E;
	int SSO = 0;
	struct Speed Position = {0, 0};

	Position.Master = QEIPositionGet(QEI0_BASE);
	Position.Slave = QEIPositionGet(QEI1_BASE);
		int PositionDiff;
		PositionDiff = Position.Master - Position.Slave;
		if(PositionDiff > 1000)
		{
			PositionDiff = PositionDiff/3;
		}
		if(PositionDiff < -1000)
		{
			PositionDiff = PositionDiff/3;
		}
		QEIPositionSet(QEI0_BASE, 1500);
		QEIPositionSet(QEI1_BASE, (1500 - PositionDiff));


	NewSpeed->Slave = PWMPulseWidthGet(PWM0_BASE, PWM_OUT_7);
	NewSpeed->Slave = NewSpeed->Slave *2000/ui32Load +1;			// This calculates the ui8Adjust values (why am I not just passing them in?)
	NewSpeed->Master = PWMPulseWidthGet(PWM0_BASE, PWM_OUT_6);
	NewSpeed->Master = NewSpeed->Master *2000/ui32Load +1;

	//Check for Bumper at this point, Bumper on PB4
	if(GPIOPinRead(GPIO_PORTB_BASE, GPIO_PIN_4))
	{
		if (strcmp(Button, "Forward") == 0|| strcmp(Button, "Left") == 0||strcmp(Button, "Right") == 0|| strcmp(Button, "ForLeft") == 0||strcmp(Button, "ForRight") == 0)
		{
			strcpy(Button, "Stop");
		}
	}

	TargetDirection = LookupControl(Button);
	if(Uncorrected->Master != Target->Master)
		{
			if(Uncorrected->Master < Target->Master)
			{
				Uncorrected->Master = Uncorrected->Master + a;
			}else
			{
				Uncorrected->Master = Uncorrected->Master - a;
			}
		}

		if(Uncorrected->Slave != Target->Slave)
			{
				if(Uncorrected->Slave < Target->Slave)
				{
					Uncorrected->Slave = Uncorrected->Slave + a;
				}else
				{
					Uncorrected->Slave = Uncorrected->Slave - a;
				}
			}
	switch(TargetDirection)
		{
		case 0:		//Stop
			NewSpeed->Slave = 150;
			NewSpeed->Master = 150;
			break;
		case 1:		//Forward
			E = (Position.Master - Position.Slave);
			/**Integral = *Integral + E;
			if(*Integral > (KI*10))
				{
					*Integral = KI*9;
				}
			if(*Integral <-(KI*10))
				{
					*Integral = -(KI*9);
				}*/
			SSO = (E/KP) + TRIMF;// + (*Integral/KI);
			NewSpeed->Slave = Uncorrected->Master +SSO;
			NewSpeed->Master = Uncorrected->Master;
			break;
		case 2:		//Reverse
			E = (Position.Master - Position.Slave);
			/**Integral = *Integral + E;
			if(*Integral > (KI*10))
			{
				*Integral = KI*9;
			}
			if(*Integral <-(KI*10))
			{
				*Integral = -(KI*9);
			}*/
			SSO = (E/KP)-TRIMR;// + (*Integral/KI);
			NewSpeed->Slave = Uncorrected->Master +SSO;
			NewSpeed->Master = Uncorrected->Master;
			/*if(Position.Master >Position.Slave)
				{
					NewSpeed->Slave = Uncorrected->Master + SSO;
				}
			if(Position.Master < Position.Slave)
				{
					NewSpeed->Slave = Uncorrected->Master - SSO;
				}
			*/
			break;
			case 3:		//Left
			case 4:		//Right
				NewSpeed->Master = Uncorrected->Master;
				NewSpeed->Slave = Uncorrected->Slave;
				break;
			case 5:		//Forward Left
			case 6:		//Forward Right
			case 7:		//Reverse Left
			case 8:		//Reverse Right
				NewSpeed->Master = Uncorrected->Master;
				NewSpeed->Slave = Uncorrected->Slave;
				break;
			default:	//Error
				NewSpeed->Slave = 150;
				NewSpeed->Master = 150;
				break;
			}
	if(NewSpeed->Master > 185)
	{
		NewSpeed->Master = 180;
	}
	if(NewSpeed->Slave > 185)
	{
		NewSpeed->Slave = 180;
	}
	if(NewSpeed->Master < 115)
	{
		NewSpeed->Master = 120;
	}
	if(NewSpeed->Slave < 115)
	{
		NewSpeed->Slave = 120;
	}
}

void UpdateSpeed (int MasterSpeed, int SlaveSpeed, uint32_t ui32Load)
{
	volatile uint32_t ui32WidthMaster;
	volatile uint32_t ui32WidthSlave;
	//Update Right Wheel (PWM6)
	ui32WidthMaster = ((MasterSpeed * ui32Load) / 2000);
	PWMPulseWidthSet(PWM0_BASE, PWM_OUT_6, ui32WidthMaster);


	//Update Left Wheel (PWM7)
	ui32WidthSlave = ((SlaveSpeed * ui32Load) / 2000);
	PWMPulseWidthSet(PWM0_BASE, PWM_OUT_7, ui32WidthSlave);
}
//*****************************************************
//
//	ChangeButton is called when an incoming UART
//	communication is detected. It first checks that the
//	incoming string is neither empty nor the NOTVALID
//	Then it copies the incoming string into Button,
//	which is then used in the rest of the functions.
//
//****************************************************
void ChangeButton(char *Button, int *Integral)
{
	char ButtonA[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

		UART1gets(ButtonA, 16);		//This way works better
		if (strcmp(ButtonA, "NOTVALID") != 0 && strcmp(ButtonA, "") != 0)
		{
				if(TOGGLE)
					{
						if(strcmp(ButtonA, "NoButton") != 0)
						{
							if(strcmp(Button,ButtonA) == 0)
									{
										strcpy(Button, "Stop");
									}else
									{
										strcpy(Button, ButtonA);
									}
						}

					}else	//This is the Hold Method
					{
						strcpy(Button, ButtonA);
					}
				QEIPositionSet(QEI0_BASE, 1500);
				QEIPositionSet(QEI1_BASE, 1500);
				*Integral = 0;
		}
}
int main(void)
{
	struct Speed Target;
	struct Speed Uncorrected = {150, 150};
	struct Speed New;
	volatile uint32_t ui32Load;
	volatile uint32_t ui32PWMClock;

	volatile int PositionMaster;
	volatile int PositionSlave;
	int Integral = 0;
	char Button[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

		SysCtlClockSet(SYSCTL_SYSDIV_5|SYSCTL_USE_PLL|SYSCTL_OSC_MAIN|SYSCTL_XTAL_16MHZ);
		PWMInit();
		QEIInit();
		ADCInit();
		UARTInit();
		GPIOInit();
		SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOF);			//unlocks PF0  (one of the user buttons built into the Launchpad)
		HWREG(GPIO_PORTF_BASE + GPIO_O_LOCK) = GPIO_LOCK_KEY;
		HWREG(GPIO_PORTF_BASE + GPIO_O_CR) |= 0x01;
		HWREG(GPIO_PORTF_BASE + GPIO_O_LOCK) = 0;
		GPIODirModeSet(GPIO_PORTF_BASE, GPIO_PIN_4|GPIO_PIN_0, GPIO_DIR_MODE_IN);
		GPIOPadConfigSet(GPIO_PORTF_BASE, GPIO_PIN_4|GPIO_PIN_0, GPIO_STRENGTH_2MA, GPIO_PIN_TYPE_STD_WPU);
		ui32PWMClock = SysCtlClockGet() / 64;
		ui32Load = (ui32PWMClock / PWM_FREQUENCY) - 1;
		New.Master = 150;
		New.Slave = 150;
		strcpy(Button, "Stop");									// Start the stander in Stop mode
		ChangeSpeed(Button, ui32Load, &Target);

		while(1)
		{

			if(UART1RxBytesAvail() != 0)
			{
				if(UART1Peek('\r') > 0)		//do not call changebutton until a complete string is present in the UART receive buffer
				{
					ChangeButton(Button, &Integral);
				}
					ChangeSpeed(Button, ui32Load, &Target);
					SysCtlDelay(10);
				}

				if(GPIOPinRead(GPIO_PORTF_BASE,GPIO_PIN_0)==0x00)
				{

					strcpy(Button,"Stop");		//Emergency stop button PF0 on the launchpad
					ChangeSpeed(Button, ui32Load, &Target);
					SysCtlDelay(10);
				}
			WheelMonitorLoop (&Target, &New, &Uncorrected, ui32Load, Button, &Integral);
			UpdateSpeed(New.Master, New.Slave, ui32Load);
			PositionMaster = QEIPositionGet(QEI0_BASE);
			PositionSlave = QEIPositionGet(QEI1_BASE);
			SysCtlDelay(100);
		}
}
